package es12;

/**
 * C.d.L. Informatica
 * Sapienza Universita' di Roma
 * Corso di Metodologie di Programmazione
 * 
 * Prova Scritta del 17/06/2009
 * 
 * Esercizio 2
 * 
 * Numeri complessi implementati usando Scalare
 * 
 * @author Stefano Guerrini 
 * @version Aprile 2009
 */
public class Complex extends Scalare<FCoppia>
{
    /**
     * Costruttore che restituisce il valore complesso zero
     * 
     * Usa direttamente il costruttore definito nella classe astratta
     */
    public Complex() {};

    /**
     * Costruttore che restituisce un nuovo scalare clonando il valore dato
     * 
     * Usa il corrispondente costruttore definito nella classe astratta
     * 
     * @param fc    coppia con i valori della parte reale e immaginaria
     */
    public Complex(FCoppia fc) { super(fc); };

    /**
     * Restituisce un nuovo oggetto complesso pari a zero
     * 
     * @return      un nuovo complesso con parte reale e immaginaria pari a 0
     */
    FCoppia zero() { return new FCoppia(); }
    
    
    /**
     * Verifica se il valore del numero complesso e' uguale a zero
     * Ovvero se sia la parte reale che quella immaginaria sono pari a 0
     * 
     * @return      true se uguale a zero, false altrimenti
     */
    public boolean isZero() {
        return (get().fst() == 0 && get().snd() == 0);
    }

    /**
     * Metodo per la clonazione di un valore
     * 
     * @param fc    la coppia di valori da clonare
     * @return      una nuova coppia di valori
     */    
    FCoppia cloneVal(FCoppia fc) { return (FCoppia)fc.clone(); };
    
    /**
     * Somma di numeri complessi
     * 
     * @param s     il secondo addendo
     * @return      un nuovo complesso con il risultato della somma
     */
    public Scalare<FCoppia> sum(Scalare<FCoppia> s) {
        FCoppia fc = new FCoppia(get().fst() + s.get().fst(),
                                    get().snd() + s.get().snd());
        return new Complex(fc);
    }
    
    /**
     * Negativo
     * 
     * @return      un nuovo scalare con il l'opposto dello scalare
     */
     public Scalare<FCoppia> neg() {
        FCoppia fc = new FCoppia(-get().fst(), -get().snd());
        return new Complex(fc);
    }

    /**
     * Moltiplicazione di numeri complessi
     * 
     * @param s     il secondo fattore
     * @return      un nuovo complesso con il risultato del prodotto
     */
     public Scalare<FCoppia> mult(Scalare<FCoppia> s) {
        float x = get().fst() * s.get().fst()
                    - (get().snd() * s.get().snd());
        float y = get().fst() * s.get().snd()
                    + (get().snd() * s.get().fst());
        FCoppia fc = new FCoppia(x,y);
        return new Complex(fc);
    }
    
    /**
     * Inverso di un numero complesso
     * 
     * @return      un nuovo scalare con l'inverso dello scalare
     */
    public Scalare<FCoppia> inv() {
        float x = get().fst();
        float y = get().snd();
        float mod = (float) Math.sqrt(x*x+y*y); 
        // ... oppure usando direttamente il metodo hypot
        // float mod = (float) Math.hypot(x,y);
        FCoppia fc = new FCoppia(x/mod, -y/mod);
        return new Complex(fc);  
    }

    /**
     * Metodo per la crezione di un nuovo numero complesso dati i suoi valori
     * 
     * @param fc    coppia dei valori reale e immaginario
     * @return      un nuovo scalare contenente la coppia data
     */
    Scalare<FCoppia> myNew(FCoppia fc) {
        return new Complex(fc);
    }
    
    /**
     * Sovrascrittura del metodo equals
     * 
     * @param o     oggetto da confrontare con l'istanza
     * @return      true se l'oggetto e' di di tipo Complex (o un sottotipo)
     *              e contiene le stesse parti reale e immaginaria dell'istanza
     */
    public boolean equals(Object o) {
        if (!(o instanceof Complex)) {
            return false;
        }
        Complex c = (Complex) o;
        return (c.get().fst() == get().fst() && c.get().snd() == get().snd());
    }
    /**
     * Sovrascrittura di hashCode
     * 
     * @return      il valore dell'hashCode, calcolato prendendo l'intero
     *              corrispondente alla rappresentazione bit a bit della
     *              somma della parte reale e della parte immaginaria
     */
    public int hashCode() {
        return (Float.floatToIntBits(get().fst() + get().snd()));
    }
}
