package es12;

/**
 * C.d.L. Informatica
 * Sapienza Universita' di Roma
 * Corso di Metodologie di Programmazione
 * 
 * Prova Scritta del 17/06/2009
 * 
 * Esercizio 1 - 2 parte
 * 
 * Implementazione generica di una classe astratta Scalare
 * su di un tipo di valore dato
 * 
 * @author Stefano Guerrini 
 * @version Aprile 2009
 */
public abstract class Scalare<T>
{
    final private T val;  // valore dello scalare

    //
    // Metodi astratti
    //
    
    /**
     * Metodo per la clonazione di un valore
     * 
     * @param t     il valore da clonare
     * @return      una nuova copia del valore
     */
    abstract T cloneVal(T t);
    
    /**
     * Somma
     * 
     * @param s     il secondo addendo
     * @return      un nuovo scalare con il risultato della somma
     */
    public abstract Scalare<T> sum(Scalare<T> s);
    
    /**
     * Negativo
     * 
     * @return      un nuovo scalare con il l'opposto
     */
    public abstract Scalare<T> neg();

    /**
     * Moltiplicazione
     * 
     * @param s     il secondo fattore
     * @return      un nuovo scalare con il risultato del prodotto
     */
    public abstract Scalare<T> mult(Scalare<T> s);
    
    /**
     * Inverso
     * 
     * @return      un nuovo scalare con l'inverso dello scalare
     */
    public abstract Scalare<T> inv();

    /**
     * Metodo per la crezione di un nuovo scalare
     * 
     * @param t     valore del nuovo scalare
     * @return      un nuovo scalare contenente il valore dato
     */
    abstract Scalare<T> myNew(T t);

    //
    // Costruttori
    //
    
    /**
     * Costruisce uno scalare di valore zero
     */
    public Scalare() { val = (T)zero(); }

    /**
     * Costruisce uno scalare clonando il valore dato
     * 
     * @param t     valore dello scalare
     */
    public Scalare(T t) { val = (T)cloneVal(t); }

    //
    // Metodi implementati
    // 
    
    /**
     * Restituisce lo zero del tipo scalare 
     * Deve essere ridefinito dalle classi concrete che
     * intendono usarlo, altrimenti solleva l'eccezione
     * UndefinedZeroScalarException
     * 
     * @return      uno scalare pari a zero
     */
    T zero() {
        throw new UndefinedZeroScalarException();
    }
    
    /**
     * Restituisce il valore dello scalare
     * 
     * @return      il valore dello scalare
     */
    public T get() { return val; }
    
    /**
     * Rende clone public
     * 
     * @return      il nuovo oggetto clonato.
     *              Utilizza un metodo astratta myNew
     *              da implementare nelle classi concrete
     */
    public Object clone() { return myNew(cloneVal(val)); }
    
    /**
     * Verifica se il valore dello scalare e' uguale a zero
     * Deve essere ridefinito dalle classi concrete che
     * intendono usarlo, altrimenti solleva l'eccezione
     * UndefinedZeroScalarException
     * 
     * @return      true se uguale a zero, false altrimenti
     */
    public boolean isZero() {
        throw new UndefinedZeroScalarException();
    }

    /**
     * Differenza
     * 
     * @param s     il sottraendo
     * @return      un nuovo scalare con il risultato della differenza
     */
    public Scalare<T> diff(Scalare<T> s) { return this.sum(s.neg()); }
    
    /**
     * Divisione
     * 
     * @param s     il divisore
     * @return      un nuovo scalare con il risultato della divisione
     */
    public Scalare<T> div(Scalare<T> s) {
        if (s.isZero())
            throw new DivByZeroScalarException();
        return this.mult(s.inv());
    }
}
